from pytube import Playlist as PytubePlaylist

from pytube.helpers import safe_filename

class Playlist:

    def __init__(self, url = None):
        self.__url = url

        if self.__url is not None:
            self.__create_playlist()

    @property
    def url(self):  
        return self.__url
    
    @url.setter
    def url(self, url):
        self.__url = url

        if url is not None:
            self.__create_playlist()

    @property
    def title(self):
        return self.__playlist.title

    @property
    def size(self):
        return self.__playlist.length

    @property
    def directory_name(self):
        # Create safe_filename from playlist title
        directory_name = safe_filename(self.__playlist.title)

        # Remove trailing characters from beginning and end of playlist_dir
        directory_name = directory_name.lstrip().rstrip()

        return directory_name
    
    def __create_playlist(self):
        self.__playlist = PytubePlaylist(self.__url)

    def __iter__(self):
        return self.__playlist.video_urls.__iter__()

    def __next__(self):
        return self.__playlist.video_urls.__next__



from warnings import catch_warnings

from sys import exit as sys_exit

with catch_warnings(record=True) as w:
    from pydub import AudioSegment

    if bool(w):
        warning_message = str(w[-1].message)

        if 'ffmpeg' in warning_message and 'avconv' in warning_message:
            print('[Warning] You need to install ffmpeg package in your system.')
            print('For more information, check our repository:')
            print('https://github.com/giovcandido/sedona')

from os import path, mkdir

class Converter:

    def __init__(self, path):
        self.__path = path
    
    @property
    def path(self):  
        return self.__path
    
    @path.setter
    def path(self, path):
        self.__path = path

    def convert_audio_stream(self, title, playlist_dir = None):
        audio_stream = AudioSegment.from_file(self.__path)

        sedona_dir = path.expanduser('~')
        sedona_dir = path.join(sedona_dir, 'SedonaMP3')

        if not path.exists(sedona_dir):
            mkdir(sedona_dir)

        # Creating playlist folder if exists
        if playlist_dir:
            sedona_dir = path.join(sedona_dir, playlist_dir)
            if not path.exists(sedona_dir):
                mkdir(sedona_dir)

        output_file = path.join(sedona_dir, title) + '.mp3'

        audio_stream.export(output_file, format="mp3", bitrate="256k")



from argparse import ArgumentParser

from colorama import init, Fore

from pyfiglet import Figlet

from shutil import get_terminal_size

from .version import __version__
from .description import __description__

def parse_arguments():
    init(autoreset=True)

    figlet = Figlet(font='standard')

    print(Fore.LIGHTRED_EX + figlet.renderText('Sedona'))

    parser =  ArgumentParser(description=__description__)

    # Add the URL argument, for a video/playlist (one URL) or a text file (multiple URL's)
    parser.add_argument('url', metavar='url', type=str, help='video/playlist url or text file with urls')

    # Set version text
    version_text = '%(prog)s {version}'.format(version=__version__)
    
    parser.add_argument('-v', '--version', action='version', version=version_text, help='show program version')

    return parser.parse_args()

def on_download_progress(stream, chunk, bytes_remaining):
    # Get total size of the file
    file_size = stream.filesize
    
    # Calculate bytes downloaded so far
    bytes_received = file_size - bytes_remaining
    
    # Calculate progress bar width
    width = int(get_terminal_size().columns * 0.55)

    # Calculate how much of the bar is filled
    filled = int(round(width * bytes_received / float(file_size)))
    
    # Set progress bar text
    progress_bar = '=' * filled + '.' * (width - filled)

    # Calculte downloaded percentage
    percentage = round(100.0 * bytes_received / float(file_size), 1)
    
    # Print progress bar
    print(Fore.LIGHTGREEN_EX + '[%s] %.2f%%' % (progress_bar, percentage), end='\r')

def on_download_complete(stream, file_path):
    print()

def print_arrow_message(message, prefix = None):
    message = f'{Fore.LIGHTCYAN_EX}=> {message}'
    
    if prefix:
        message = prefix + message

    print(message)

def print_separator():
    separator = '\n' + get_terminal_size().columns * '-' + '\n'

    print(Fore.LIGHTCYAN_EX + separator)



from sys import exit as sys_exit

from .cli import parse_arguments, print_separator, print_arrow_message

from .Video import Video
from .Playlist import Playlist
from .Converter import Converter

def main():
    # Get user arguments from sedona command
    args = parse_arguments()

    # Create a list to store the urls
    urls = list()

    # Check if it's necessary to read urls from text file
    if args.url.endswith('.txt'):
        # If it is, then read it from file
        urls = read_from_file(args.url)
    else:
        # Otherwise, it's a single url, store it in the list
        urls.append(args.url)
    
    # Walk through urls list downloading and converting the corresponding videos
    size_urls = len(urls)

    for position, url in enumerate(urls):
        # Check if the url is a video
        if "playlist" not in url:
            # If so, let's download and convert it
            handle_video(url, position, size_urls)
        else:
            # Otherwise, let's download and convert a playlist
            handle_playlist(url, position, size_urls)

def read_from_file(filename):
    urls = None

    try:
        with open(filename) as f:
            urls = f.read().splitlines()
    except FileNotFoundError as err:
        print(err)
        sys_exit(1)
    
    return urls

def handle_video(url, position, size_urls):
    try:
        # Create video instance with the current url
        video = Video(url)

        # Show some informations about track
        print(f'Video title: {video.title}')
        print(f'Duration: {video.duration}')
        print(f'YouTube channel: {video.channel}\n')
    
        print('Downloading audio-only video...')

        # Download audio only video to the tmp directory
        video_path = video.download_audio_stream()
        
        # Create converter instance with the video path
        converter = Converter(video_path)

        print('Converting downloaded video to mp3...')

        # Get downloaded video and convert it to mp3, store it in home/SedonaMP3
        converter.convert_audio_stream(video.filename) # default bitrate = 256kbps

        print('Done! File saved to your home directory.')

        # Don't print the separator in the last element
        if position < size_urls - 1:
            print_separator()

    except Exception as err:
        print(err)
        sys_exit(1)

def handle_playlist(url, position, size_urls):
    try:
        # Create playlist instance with current url
        playlist = Playlist(url)
        
        print_arrow_message(f'Current playlist is "{playlist.title}".\n')

        # Walk through all urls of the playlist                
        for number, video_url in enumerate(playlist):
            # Create video instance with the curl
            video = Video(video_url)
            
            # Show some informations about track
            print(f'Track number: {str((number + 1))}')
            print(f'Video title: {video.title}')
            print(f'Duration: {video.duration}')
            print(f'YouTube channel: {video.channel}\n')

            print('Downloading audio-only video...')
            
            # Download audio only video to the tmp directory
            video_path = video.download_audio_stream()

            # Create converter instance with the video path
            converter = Converter(video_path)

            print('Converting downloaded video to mp3...')
            
            # Add track number to filename
            filename = str((number + 1)) + "." + " " + video.filename

            # Get downloaded video and convert it to mp3, store it in home/SedonaMP3
            converter.convert_audio_stream(filename, playlist.directory_name) # default bitrate = 256kbps

            print('Done! File saved to your home directory.')

            # Don't print the separator in the last element
            if number < playlist.size - 1:
                print_separator()

        print_arrow_message(f'Playlist "{playlist.title}" downloaded successfully.', '\n')

        # Don't print the separator in the last element
        if position < size_urls - 1:
            print_separator()

    except Exception as err:
        print(err)
        sys_exit(1)



from pytube import YouTube

from tempfile import gettempdir

from os import path, mkdir, remove

from .cli import on_download_progress, on_download_complete

class Video:

    def __init__(self, url = None):
        self.__url = url

        if self.__url is not None:
            self.__create_audio_stream()

    @property
    def url(self):  
        return self.__url
    
    @url.setter
    def url(self, url):
        self.__url = url

        if url is not None:
            self.__create_audio_stream()

    @property
    def title(self):
        return self.__audio_stream.title
    
    @property
    def duration(self):
        m, s = divmod(self.__youtube.length, 60)
        h, m = divmod(m, 60)

        if(h > 0):
            time = str(h) + ':' + str(m) + ':' + str(s)
        else:
            time = str(m) + ':' + str(s)

        return time
    
    @property
    def channel(self):
        return self.__youtube.author
    
    @property
    def filename(self):
        basename = self.__audio_stream.default_filename

        return path.splitext(basename)[0]
    
    def __create_audio_stream(self):
        self.__youtube = YouTube(self.__url)

        self.__youtube.register_on_progress_callback(on_download_progress)
        self.__youtube.register_on_complete_callback(on_download_complete)

        self.__audio_stream = self.__youtube.streams.get_audio_only()

    def download_audio_stream(self):
        temp_dir = gettempdir()

        temp_dir = path.join(temp_dir, 'Sedona')

        if not path.exists(temp_dir):
            mkdir(temp_dir)

        filename = self.__audio_stream.default_filename
        
        output_file = path.join(temp_dir, filename)
        
        if path.exists(output_file):
            remove(output_file)

        self.__audio_stream.download(output_path=temp_dir, filename=filename)

        return output_file



from setuptools import setup, find_packages

from os.path import join

# Load version
exec(open(join('sedona', 'version.py')).read())

# Load long description 
with open("README.md", encoding='utf-8') as f:
    long_description = f.read()

# Load requirements
with open("requirements.txt", encoding='utf-8') as f:
    requirements = f.read()

# Configure setup 
setup(
    name = 'sedona',
    version = __version__, # type: ignore
    author = 'Giovani Candido, Davi Neves',
    author_email = 'giovcandido@outlook.com, davivirgula@gmail.com',
    license = 'GNU General Public License v3.0',
    description = 'Sedona is a free youtube mp3 downloader made with Python.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/giovcandido/sedona',
    packages = find_packages(),
    install_requires = [requirements],
    python_requires = '>=3.8',
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.8",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    entry_points = '''
        [console_scripts]
        sedona=sedona.__main__:main
    '''
)
